from __future__ import division, absolute_import
import numpy as np

cimport numpy as np

cdef extern from "force.h":
    void compute_accelerations(int len_particles, const double *xpos, const double *ypos, double *xacc, double *yacc,
                               double *distances, double* forces, double box_length)
    void compute_sd(int len_particles, const double *xpos, const double *ypos, double *energy, double *xforce,
                    double *yforce, double box_length)
    void compute_energy_and_force(int len_particles, const double *xpos, const double *ypos, double *energy,
                                  double *xforce, double *yforce, double *xforcedash, double *yforcedash,
                                  double box_length)
    void compute_force(int len_particles, const double *xpos, const double *ypos, double *xforce, double *yforce,
                       double box_length)
    void scale_velocities(int len_particles, double *xvel, double *yvel, double average_temp, double tempature)


DTYPE = np.float64
ctypedef np.float64_t DTYPE_t


def compute_forces(particles, system):
    cdef int len_particles = particles.size
    cdef double box_length = system.box_length
    cdef np.ndarray[DTYPE_t, ndim=1] xpos = np.zeros(particles.size)
    cdef np.ndarray[DTYPE_t, ndim=1] ypos = np.zeros(particles.size)
    cdef np.ndarray[DTYPE_t, ndim=1] xacc = np.zeros(particles.size)
    cdef np.ndarray[DTYPE_t, ndim=1] yacc = np.zeros(particles.size)
    cdef np.ndarray[DTYPE_t, ndim=1] distances = np.zeros(len(system.distances))
    cdef np.ndarray[DTYPE_t, ndim=1] forces = np.zeros(len(system.forces))

    for i in range(0, particles.size):
        xpos[i] = particles[i].xpos
        ypos[i] = particles[i].ypos
        xacc[i] = particles[i].xacc
        yacc[i] = particles[i].yacc

    compute_accelerations(len_particles, <const double*>xpos.data, <const double*>ypos.data, <double*>xacc.data,
                          <double*>yacc.data, <double*>distances.data, <double*>forces.data, box_length)

    for i in range(0, particles.size):
        particles[i].xpos = xpos[i]
        particles[i].ypos = ypos[i]
        particles[i].xacc = xacc[i]
        particles[i].yacc = yacc[i]

    system.distances = distances
    system.forces = forces

    return particles, system

def calculate_sd(particles, system):
    cdef int len_particles = particles.size
    cdef np.ndarray[DTYPE_t, ndim=1] xpos = np.zeros(particles.size)
    cdef np.ndarray[DTYPE_t, ndim=1] ypos = np.zeros(particles.size)
    cdef np.ndarray[DTYPE_t, ndim=1] energy = np.zeros(particles.size)
    cdef np.ndarray[DTYPE_t, ndim=1] xforce = np.zeros(particles.size)
    cdef np.ndarray[DTYPE_t, ndim=1] yforce = np.zeros(particles.size)
    cdef double box_length = system.box_length

    for i in range(0, particles.size):
        xpos[i] = particles[i].xpos
        ypos[i] = particles[i].ypos

    compute_sd(len_particles, <const double*> xpos.data, <const double*> ypos.data, <double*> energy.data,
                             <double*>xforce.data, <double*> yforce.data, box_length)

    for i in range(0, particles.size):
        particles[i].energy = energy[i]
        particles[i].xforce = xforce[i]
        particles[i].yforce = yforce[i]

    return particles

def calculate_energy_and_force(particles, system):
    cdef int len_particles = particles.size
    cdef np.ndarray[DTYPE_t, ndim=1] xpos = np.zeros(particles.size)
    cdef np.ndarray[DTYPE_t, ndim=1] ypos = np.zeros(particles.size)
    cdef np.ndarray[DTYPE_t, ndim=1] energy = np.zeros(particles.size)
    cdef np.ndarray[DTYPE_t, ndim=1] xforce = np.zeros(particles.size)
    cdef np.ndarray[DTYPE_t, ndim=1] yforce = np.zeros(particles.size)
    cdef np.ndarray[DTYPE_t, ndim=1] xforcedash = np.zeros(particles.size)
    cdef np.ndarray[DTYPE_t, ndim=1] yforcedash = np.zeros(particles.size)
    cdef double box_length = system.box_length

    for i in range(0, particles.size):
        xpos[i] = particles[i].xpos
        ypos[i] = particles[i].ypos


    compute_energy_and_force(len_particles, <const double*> xpos.data, <const double*> ypos.data, <double*> energy.data,
                             <double*>xforce.data, <double*> yforce.data, <double*> xforcedash.data,
                             <double*> yforcedash.data, box_length)

    for i in range(0, particles.size):
        particles[i].energy = energy[i]
        particles[i].xforce = xforce[i]
        particles[i].yforce = yforce[i]
        particles[i].xforcedash = xforcedash[i]
        particles[i].yforcedash = yforcedash[i]

    return particles

def get_forces(particles, system):
    cdef int len_particles = particles.size
    cdef np.ndarray[DTYPE_t, ndim=1] xpos = np.zeros(particles.size)
    cdef np.ndarray[DTYPE_t, ndim=1] ypos = np.zeros(particles.size)
    cdef np.ndarray[DTYPE_t, ndim=1] xforce = np.zeros(particles.size)
    cdef np.ndarray[DTYPE_t, ndim=1] yforce = np.zeros(particles.size)
    cdef double box_length = system.box_length

    for i in range(0, particles.size):
        xpos[i] = particles[i].xpos
        ypos[i] = particles[i].ypos


    compute_force(len_particles, <const double*> xpos.data, <const double*> ypos.data, <double*>xforce.data,
                  <double*> yforce.data, box_length)

    return xforce, yforce


def heat_bath(particles, system):
    cdef int len_particles = particles.size
    cdef np.ndarray[DTYPE_t, ndim=1] xvel = np.zeros(particles.size)
    cdef np.ndarray[DTYPE_t, ndim=1] yvel = np.zeros(particles.size)
    cdef double average_temp = np.average(system.temp_array)
    cdef double temperature = system.temperature

    for i in range(0, particles.size):
        xvel[i] = particles[i].xvel
        yvel[i] = particles[i].yvel

    scale_velocities(len_particles, <double*>xvel.data, <double*>yvel.data, average_temp, temperature)

    for i in range(0, particles.size):
        particles[i].xvel = xvel[i]
        particles[i].yvel = yvel[i]

    return particles


import numpy as np
from pylj import force, sample, util


def clear_accelerations(particles):
    """Reset the accelerations to 0.

    Parameters
    ----------
    particles: Particle array
        The particles in the system.

    Returns
    -------
    Particle array
        The particles in the system now with zero acceleration.
    """
    for i in range(0, len(particles)):
        particles[i].xacc = 0
        particles[i].yacc = 0
    return particles


def compute_acceleration(particles, system):
    """A pure python method for the calculation of forces and accelerations on each atom.

    Parameters
    ----------
    particles: Particle array
        The particles in the system.
    system: System
        System parameters.

    Returns
    -------
    Particle array
        The particles with the accelerations calculated.
    System
        System parameters, where the distances array has been updated.
    """
    system.distances = []
    particles = clear_accelerations(particles)
    for i in range(0, len(particles) - 1):
        for j in range(i + 1, len(particles)):
            dx = particles[i].xpos - particles[j].xpos
            dy = particles[i].ypos - particles[j].ypos
            dx = util.pbc_correction(dx, system.box_length)
            dy = util.pbc_correction(dy, system.box_length)
            dr = np.sqrt(dx * dx + dy * dy)
            system.distances = np.append(system.distances, dr)
            f = 48 * np.power(dr, -13.) - 24 * np.power(dr, -7.)
            particles[i].xacc += f * dx / dr
            particles[i].yacc += f * dy / dr
            particles[j].xacc -= f * dx / dr
            particles[j].yacc -= f * dy / dr
    return particles, system


def reset_histogram(system):
    """Reset the velocity histogram.

    Parameters
    ----------
    system: System
        System parameters.

    Returns
    -------
    System
        System parameters with the velocity bins and temperature summation set to zero.
    """
    for i in range(0, len(system.vel_bins)):
        system.vel_bins[i] = 0
    system.temp_sum = 0
    system.step0 = system.step
    return system


def initialise(number_of_particles, temperature, timestep_length, arrangement):
    """Initial particle positions (simple square arrangment), velocities and get initial forces/accelerations.

    Parameters
    ----------
    number_of_particles: int
        Number of particles in the system.
    temperature: float
        Temperature of the system.
    arrangement: func
        A function to set the arrangement of the particles.

    Returns
    -------
    Particle array
        An array containing the initial particle positions, velocities and accelerations set.
    System
        System information.
    """
    system = util.System(number_of_particles, temperature, 16., timestep_length)
    particles = arrangement(system)
    v = np.sqrt(2 * system.temperature)
    for i in range(0, system.number_of_particles):
        theta = 2 * np.pi * np.random.randn()
        particles[i].xvel = v * np.cos(theta)
        particles[i].yvel = v * np.sin(theta)
    particles, system = force.compute_forces(particles, system)
    system = reset_histogram(system)
    return particles, system


def initialise_from_em(particles, system, temperature, timestep_length):
    system.temperature = temperature
    system.timestep_length = timestep_length
    v = np.sqrt(2 * system.temperature)
    for i in range(0, system.number_of_particles):
        theta = 2 * np.pi * np.random.randn()
        particles[i].xvel = v * np.cos(theta)
        particles[i].yvel = v * np.sin(theta)
    particles, system = force.compute_forces(particles, system)
    system = reset_histogram(system)
    return particles, system


def update_pos_verlet(particle, system):
    """Update the positions of a given particle based on the integrator.

    Parameters
    ----------
    particle: Particle
        A particle in the system.
    system: System
        Whole system information.

    Returns
    -------
    Particle
        Particle with updated positions.
    """
    particle.xpos = particle.xpos - particle.xpos_prev + particle.xacc * system.timestep_length * system.timestep_length
    particle.ypos = particle.ypos - particle.ypos_prev + particle.yacc * system.timestep_length * system.timestep_length
    particle.xpos = particle.xpos % system.box_length
    particle.ypos = particle.ypos % system.box_length
    return particle


def update_pos_vv(particle, system):
    """Update the positions of a given particle based on the integrator.

    Parameters
    ----------
    particle: Particle
        A particle in the system.
    system: System
        Whole system information.

    Returns
    -------
    Particle
        Particle with updated positions.
    """
    particle.xpos += particle.xvel * system.timestep_length + 0.5 * particle.xacc * \
                                                                      system.timestep_length * system.timestep_length
    particle.ypos += particle.yvel * system.timestep_length + 0.5 * particle.yacc * \
                                                                      system.timestep_length * system.timestep_length
    particle.xpos = particle.xpos % system.box_length
    particle.ypos = particle.ypos % system.box_length
    return particle


def update_velocities_verlet(particle, system):
    """Update the velocities of a given particles based on the acceleration.

    Parameters
    ----------
    particle: Particle
        A particle in the system.
    system: System
        Whole system information.

    Returns
    -------
    Particle
        Particle with updated velocities.
    """
    particle.xvel = (particle.xpos - particle.xpos_prev) / (2 * system.timestep_length)
    particle.yvel = (particle.ypos - particle.ypos_prev) / (2 * system.timestep_length)
    return particle


def update_velocities_vv(particle, system):
    """Update the velocities of a given particles based on the acceleration.

    Parameters
    ----------
    particle: Particle
        A particle in the system.
    system: System
        Whole system information.

    Returns
    -------
    Particle
        Particle with updated velocities.
    """
    particle.xvel += 0.5 * particle.xacc * system.timestep_length
    particle.yvel += 0.5 * particle.yacc * system.timestep_length
    return particle


def update_velocity_bins(particle, system):
    """Updates the velocity bins. It is not clear if this is completely required if the velocity bins are not
    being plotted.

    Parameters
    ----------
    particle: Particle
        A particle in the system.
    system: System
        Whole system information.

    Returns
    -------
    System
        Whole system information with the velocity bin for particle updated.
    float
        The total velocity for a given particle.
    """
    v = np.sqrt(particle.xvel * particle.xvel + particle.yvel * particle.yvel)
    bin_s = int(len(system.vel_bins) * v / system.max_vel)
    if bin_s < 0:
        bin_s = 0
    if bin_s >= len(system.vel_bins):
        bin_s = len(system.vel_bins) - 1
    system.vel_bins[bin_s] += 1
    return system, v


def velocity_verlet(particles, system):
    """Update the positions, velocities, get temperature and pressure.

    Parameters
    ----------
    particles: Particle array
        All particles in the system.
    system: System
        Whole system information.

    Returns
    -------
    Particle array:
        Particles with updated positions and velocities.
    System:
        Whole system information with new temperature and pressure.
    """
    system = reset_histogram(system)
    system.step += 1
    for i in range(0, system.number_of_particles):
        position_store = [particles[i].xpos, particles[i].ypos]
        particles[i] = update_pos_vv(particles[i], system)
        particles[i] = update_velocities_vv(particles[i], system)
        particles[i].xpos_prev = position_store[0]
        particles[i].ypos_prev = position_store[1]
    particles, system = util.calculate_temperature(particles, system)
    system = util.calculate_pressure(system)
    system.force_array.append(np.sum(system.forces))
    return particles, system


def verlet(particles, system):
    """Update the positions, velocities, get temperature and pressure.

    Parameters
    ----------
    particles: Particle array
        All particles in the system.
    system: System
        Whole system information.

    Returns
    -------
    Particle array:
        Particles with updated positions and velocities.
    System:
        Whole system information with new temperature and pressure.
    """
    system = reset_histogram(system)
    system.step += 1
    for i in range(0, system.number_of_particles):
        particles[i] = update_pos_verlet(particles[i], system)
        particles[i] = update_velocities_verlet(particles[i], system)
    particles, system = util.calculate_temperature(particles, system)
    system = util.calculate_pressure(system)
    return particles, system


def scale_velocities(particles, system):
    """Pure python for the velocity rescaling.

    Parameters
    ----------
    particles: Particle array
        All particles in the system.
    system: System
        Whole system information.

    Returns
    -------
    Particle array:
        All particles with velocities rescaled.
    """
    for i in range(0, len(particles)):
        particles[i].xvel *= 1 / np.average(system.temp_array)
        particles[i].yvel *= 1 / np.average(system.temp_array)
    return particles


import numpy as np
from pylj import md, force


class System:
    """Whole system.

    This class stores a large amount of information about the job that is being run.

    Parameters
    ----------
    number_of_particles: int
        Number of particles to be simulated.
    kinetic_energy: float
        The initial kinetic energy of the system, this same value is used to reweight the velocities.
    box_length: float
        Size of the simulation cell.
    timestep_length: float
        Duration of timestep in MD integration.
    max_vel: float, optional
        Maximum velocity allowed in the velocity rebinning stage.
    threshold: float, optional
        The energy change threshold for running an energy minimisation.
    max_steps: int, optional
        The maximum number of steps in the energy minimisation process.
    """
    def __init__(self, number_of_particles, temperature, box_length, timestep_length,
                 max_vel=4, threshold=1e-20, max_steps=1000):
        self.number_of_particles = number_of_particles
        self.temperature = temperature
        self.box_length = box_length
        self.timestep_length = timestep_length
        self.temp_sum = 0.
        self.vel_bins = np.zeros(500)
        self.max_vel = max_vel
        self.step = 0
        self.step0 = 0
        pairs = (self.number_of_particles-1)*self.number_of_particles/2
        self.distances = np.zeros(int(pairs))
        self.forces = np.zeros(int(pairs))
        self.temp_array = []
        self.press_array = [0]
        self.bin_width = 0.1
        self.time = 0
        self.total_force = np.zeros(3)
        self.old_total_force = np.zeros(3)
        self.threshold = threshold
        self.max_steps = max_steps
        self.force_array = []


class Particle:
    """Particles

    This stores particle relevant information.

    Parameters
    ----------
    xpos: float
        Position in the x-axis.
    ypos: float
        Position in the y-axis.
    xvel: float
        Velocity in the x-axis.
    yvel: float
        Velocity in the y-axis.
    xacc: float
        Acceleration in the x-axis.
    yacc: float
        Acceleration in the y-axis.
    """
    def __init__(self, xpos, ypos, xvel, yvel, xacc, yacc):
        self.xpos = xpos
        self.ypos = ypos
        self.xvel = xvel
        self.yvel = yvel
        self.xacc = xacc
        self.yacc = yacc
        self.xpos_prev = 0.
        self.ypos_prev = 0.
        self.energy = 0.
        self.xforce = 0.
        self.yforce = 0.
        self.xforcedash = 0.
        self.yforcedash = 0.


def pbc_correction(d, l):
    """Test and correct for the periodic boundary condition.

    Parameters
    ----------
    d: float
        Particle position.
    l: float
        Box vector.

    Returns
    -------
    float
        Corrected particle position."""
    if np.abs(d) > 0.5 * l:
        d *= 1 - l / np.abs(d)
    return d

def calculate_pressure(system):
    """Calculates the instantaneous pressure of the system.

    Parameters
    ----------
    system: System
        Whole system information

    Returns
    -------
    System
        System with updated press_array to include newest instantaneous pressure.
    """
    w = (-1. / 3. * np.sum(system.distances * -1 * system.forces))
    system.press_array.append(system.number_of_particles * system.temperature + w)
    return system

def calculate_temperature(particles, system):
    """Determine the instantaneous temperature of the system.

    Parameters
    ----------
    particles: Particle array
        All particles in the system.
    system: System
        Whole system information.

    Returns
    -------
    Particle array:
        All particles updated with their velocities scaled.
    System:
        Whole system information with the temperature updated."""
    k = 0
    for i in range(0, system.number_of_particles):
        system, v = md.update_velocity_bins(particles[i], system)
        k += 0.5 * v * v
    system.temp_sum += k / system.number_of_particles
    temp = system.temp_sum / (system.step - system.step0)
    system.temp_array.append(temp)
    return particles, system


def set_particles_square(system):
    """Set the initial particle positions on a square lattice.

    Parameters
    ----------
    system: System
        Whole system information.

    Returns
    -------
    Particle array
        The particles with positions on a square lattice.
    """
    particles = np.array([], dtype=Particle)
    m = int(np.ceil(np.sqrt(system.number_of_particles)))
    d = system.box_length / m
    n = 0
    for i in range(0, m):
        for j in range(0, m):
            if n < system.number_of_particles:
                part = Particle((i + 0.5) * d, (j + 0.5) * d, 0, 0, 0, 0)
                particles = np.append(particles, part)
                n += 1
    return particles


def set_particles_random(system):
    """Set the initial particle positions in a random arrangement.

    Parameters
    ----------
    system: System
        Whole system information.

    Returns
    -------
    Particle array
        The particles with position set randomly.
    """
    particles = np.array([], dtype=Particle)
    for i in range(0, system.number_of_particles):
        part = Particle(np.random.uniform(0, system.box_length), np.random.uniform(0, system.box_length),
                        0, 0, 0, 0)
        particles = np.append(particles, part)
    return particles